#include	<stdio.h>
#include	<stdlib.h>
#include	<dos.h>
#include	<conio.h>
#include	<ctype.h>
#include	<string.h>

#include	"zstddef.h"
#include	"timerms.h"

#define	COLUMN1	3
#define	COLUMN2	43
#define	VALOFS	4
#define	SNDROWS	10

// Macro for determining the number of members in an array.

#define	NUMMEM( aa)		(sizeof( aa) / sizeof( *(aa)))

// Set default port address

//#define	DEFPORTADR	0x378		// LPT1
#define	DEFPORTADR	0x278		// LPT2

#define	PULSE38HZ	31400		// number of 8254 ticks for a 38hz pulse

enum SOUNDACTIONS_E
{	ACT_SET,							// Set sound value
	ACT_TGL,							// Toggle sound value
	ACT_PLS,							// Pulse sound value
	ACT_TGR							// Special 'Tailgunner' action
};

// To create a new sound type (shift register, etc):
// Add a new ST_xxx value here.
// Create new 'xxxPortVal' and 'xxxPortOld' globals.
// Write the 'readxxxBit()', 'setxxxBit()', and 'setxxxReg()' routines.
// Add the routines to the 'ReadBitTbl[]', 'SetBitTbl[]' and 'SetRegTbl[]' tables.
// Add a 'xxxPortOld = ~xxxPortVal' initialization line to the routine 'initSoundBd()'.
//
// NOTE: The order of this 'enum' is critical!  It must correspond to the order
// of the tables:  'SetRegTbl[]', 'SetBitTbl[]' and 'ReadBitTbl[]'.

enum SOUNDTYPES_E
{	ST_OUT,							// Standard Out Port
	ST_AAS,							// Armor Attack Shift register
	ST_ROS,							// Ripoff Shift register
	ST_SQ1,							// Solar Quest 8 bit shift register
	ST_SQ2,							// Solar Quest 16 bit shift register
	ST_TGS,							// Tailgunner 3 to 8 mux strobe

	ST_NUM							// number of sound types (must be last)
};

// Structure used to define a single sound.

typedef struct SOUND_S
{	char	*name;					// name of sound
	uint	action;					// default action (See SOUNDACTIONS_E)
	uint	timing;					// number of 38hz pulses used to create sound
	uint	type;						// sound type (See SOUNDTYPES_E)
	uint	bitMask;					// used to mask used bit(s) (pre-shift)
	uint	invMask;					// used to invert bit(s), if needed, for negative logic (pre-shift)
	uint	shlCount;				// shift left count, used to position bit(s) in port
} Sound_s;

// Structure used to define a sound board.

typedef struct SOUNDBD_S
{	char			*title;			// name of sound board
	uint			nsnd;				// number of sounds in array
	Sound_s		*sound;			// pointer to sounds
} SoundBd_s;

// Ports and previous port values for different sound cards.

uint		Port;								// address of LPT port
uchar		OutPortVal, OutPortOld;		// value of LPT port, and previous value of LPT port
uchar		AASPortVal, AASPortOld;		// values of the Armor Attack Shift Register
uchar		ROSPortVal, ROSPortOld;		// values of the Ripoff Shift Register
uchar		TGSPortVal, TGSPortOld;		// values of the Tailgunner 3 to 8 mux Register
uchar		SQ1PortVal, SQ1PortOld;		// values of the Solar Quest 8 bit shift register
uint		SQ2PortVal, SQ2PortOld;		// values of the Solar Quest 16 bit shift register

SoundBd_s	*Sbd;							// pointer to the attached sound board

// Setup table arrays that define each known sound board.

// Amor Attack

Sound_s AASounds[] =
{
	{ "Tank Enable",    ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 1},
	{ "Beep Enable",    ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 2},
	{ "Chopper Enable", ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 3},
	{ "Tank Fire",      ACT_PLS, 1, ST_AAS, 0x0001, 0x0001, 7},
	{ "Hi Explosion",   ACT_PLS, 1, ST_AAS, 0x0001, 0x0001, 6},
	{ "Jeep Fire",      ACT_PLS, 1, ST_AAS, 0x0001, 0x0001, 5},
	{ "Lo Explosion",   ACT_PLS, 1, ST_AAS, 0x0001, 0x0001, 4},
	{ "Tank Speed",     ACT_SET, 0, ST_AAS, 0x000F, 0x0000, 0}
};

// Rip Off

Sound_s ROSounds[] =
{
	{ "Explosion",         ACT_PLS, 0, ST_OUT, 0x0001, 0x0001, 7},
	{ "Laser",             ACT_PLS, 4, ST_OUT, 0x0001, 0x0001, 4},
	{ "Torpedo",           ACT_PLS, 3, ST_OUT, 0x0001, 0x0001, 3},
	{ "Background Speed",  ACT_SET, 0, ST_ROS, 0x0007, 0x0000, 3},
	{ "Background Enable", ACT_TGL, 0, ST_ROS, 0x0001, 0x0001, 2},
	{ "Beep",              ACT_TGL, 0, ST_ROS, 0x0001, 0x0001, 1},
	{ "Motor",             ACT_TGL, 0, ST_ROS, 0x0001, 0x0001, 0}
};

// Solar Quest

Sound_s SQSounds[] =
{
	{ "Loud Explosion", ACT_PLS, 1, ST_SQ1, 0x0001, 0x0001,  7},
	{ "Soft Explosion", ACT_PLS, 1, ST_SQ1, 0x0001, 0x0001,  6},
	{ "Thrust",         ACT_TGL, 0, ST_SQ1, 0x0001, 0x0001,  5},
	{ "Fire",           ACT_PLS, 1, ST_SQ1, 0x0001, 0x0001,  4},
	{ "Capture",        ACT_PLS, 1, ST_SQ1, 0x0001, 0x0001,  3},
	{ "Nuke",           ACT_TGL, 0, ST_SQ1, 0x0001, 0x0000,  2},
	{ "Photon",         ACT_PLS, 1, ST_SQ1, 0x0001, 0x0001,  1},
	{ "Music Enable",   ACT_TGL, 0, ST_SQ2, 0x0001, 0x0000, 15},
	{ "Music Volume",   ACT_SET, 0, ST_SQ2, 0x0007, 0x0007, 12},
	{ "Music Pitch",    ACT_SET, 0, ST_SQ2, 0x0FFF, 0x0000,  0}
};

// Space War

Sound_s SWSounds[] =
{
	{ "Explosion",    ACT_PLS, 1, ST_OUT, 0x0001, 0x0000, 0},
	{ "Fire",         ACT_PLS, 1, ST_OUT, 0x0001, 0x0000, 1},
	{ "Thrust 1",     ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 2},
	{ "Thrust 2",     ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 3},
	{ "Sound Enable", ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 4}
};

// Star Castle

Sound_s SCSounds[] =
{
	{ "Loud Explosion",    ACT_PLS, 1, ST_OUT, 0x0001, 0x0001, 1},
	{ "Soft Explosion",    ACT_PLS, 1, ST_OUT, 0x0001, 0x0001, 2},
	{ "Laser Enable",      ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 3},
	{ "Fire Ball Enable",  ACT_TGL, 0, ST_AAS, 0x0001, 0x0001, 7},
	{ "Shield Enable",     ACT_TGL, 0, ST_AAS, 0x0001, 0x0001, 6},
	{ "Star Enable",       ACT_TGL, 0, ST_AAS, 0x0001, 0x0000, 5},
	{ "Thrust Enable",     ACT_TGL, 0, ST_AAS, 0x0001, 0x0001, 4},
	{ "Background Enable", ACT_TGL, 0, ST_AAS, 0x0001, 0x0001, 3},
	{ "Background Speed",  ACT_SET, 0, ST_AAS, 0x0007, 0x0000, 0}
};

// Star Hawk

Sound_s SHSounds[] =
{
	{ "Explosion",     ACT_PLS, 0, ST_OUT, 0x0001, 0x0000, 0},
	{ "Laser 1",       ACT_PLS, 1, ST_OUT, 0x0001, 0x0000, 1},
	{ "Laser 2",       ACT_TGL, 0, ST_OUT, 0x0001, 0x0000, 2},
	{ "K On/Off",      ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 3},
	{ "Master On/Off", ACT_TGL, 0, ST_OUT, 0x0001, 0x0001, 4},
	{ "K Exit",        ACT_TGL, 0, ST_OUT, 0x0001, 0x0000, 7}
};

// Sundance

Sound_s SDSounds[] =
{
	{ "Bong",      ACT_PLS, 1, ST_OUT, 0x0001, 0x0001, 0},
	{ "Whoosh",    ACT_PLS, 0, ST_OUT, 0x0001, 0x0001, 1},
	{ "Explosion", ACT_PLS, 0, ST_OUT, 0x0001, 0x0001, 2},
	{ "Ping 1",    ACT_PLS, 1, ST_OUT, 0x0001, 0x0001, 3},
	{ "Ping 2",    ACT_PLS, 1, ST_OUT, 0x0001, 0x0001, 4},
	{ "Hatch",     ACT_PLS, 6, ST_OUT, 0x0001, 0x0001, 7}
};

// Tailgunner

Sound_s TGSounds[] =
{
	{ "Explosion",     ACT_PLS,  0, ST_TGS, 0x0001, 0x0001, 0},
	{ "Rumble",        ACT_TGL,  0, ST_TGS, 0x0001, 0x0001, 1},
	{ "Laser",         ACT_TGR,  4, ST_TGS, 0x0001, 0x0001, 2},
	{ "Shield",        ACT_TGL,  0, ST_TGS, 0x0001, 0x0001, 3},
	{ "Shield Bounce", ACT_PLS,  0, ST_TGS, 0x0001, 0x0001, 4},
	{ "Hyperspace",    ACT_PLS, 34, ST_TGS, 0x0001, 0x0001, 5},
};

// Setup a table array that defines all the known sound cards, by
// pointing to each of the above sound board definition arrays.

SoundBd_s SoundBoard[] =
{
	{ "Armor Attack", NUMMEM( AASounds), AASounds},
	{ "Rip Off",      NUMMEM( ROSounds), ROSounds},
	{ "Solar Quest",  NUMMEM( SQSounds), SQSounds},
	{ "Space War(s)", NUMMEM( SWSounds), SWSounds},
	{ "Star Castle",  NUMMEM( SCSounds), SCSounds},
	{ "Star Hawk",    NUMMEM( SHSounds), SHSounds},
	{ "Sundance",     NUMMEM( SDSounds), SDSounds},
	{ "Tailgunner",   NUMMEM( TGSounds), TGSounds}
};

/***** Standard Out port *****/

// Set the main 'out' port given in 'Port', with the value 'PortVal'

void setOutReg( void)
{
	if (OutPortVal != OutPortOld)
	{	outp( Port, OutPortVal);
		OutPortOld = OutPortVal;
	}
}

// Set outbit to value in 'val'.

void setOutBit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	OutPortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in port
	OutPortVal |= val;										// set/reset bit(s) in port
}

// Read a bit from the OUT register.

uint readOutBit( Sound_s *snd)
{
	return (((OutPortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}

/***** Armor Attack / Star Castle shift register *****/

// Set the Armor Attack shift register

void setAASReg( void)
{
	uint	ii, mask;

	if (AASPortVal != AASPortOld)
	{	OutPortVal &= 0x6E;									// reset data, clock and latch lines
		mask = AASPortVal;									// get current port value

		for (ii = 0; ii < 8; ii++)
		{
			// place bit-0 of 'mask' into OUT-7 bit

			if (mask & 0x01)
				OutPortVal |= 0x80;

			else
				OutPortVal &= 0x7F;

			outp( Port, OutPortVal);						// output bit
			OutPortVal |= 0x10;								// clock in data
			outp( Port, OutPortVal);						// update port
			OutPortVal &= 0xEF;								// reset clock
			outp( Port, OutPortVal);						// update port
			mask >>= 1;											// get next bit
		}
		OutPortVal |= 0x01;									// latch in new data
		outp( Port, OutPortVal);							// update port
		OutPortVal &= 0xFE;									// reset latch
		outp( Port, OutPortVal);							// update port
		AASPortOld = AASPortVal;							// remember new value
		
		OutPortOld = OutPortVal;							// remember new Port value
	}
}			

// Set AASbit to value in 'val'.

void setAASBit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	AASPortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in port
	AASPortVal |= val;										// set/reset bit(s) in port
}

// Read an AASbit

uint readAASBit( Sound_s *snd)
{
	return (((AASPortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}

/***** Ripoff shift register *****/

// Set the Ripoff shift register

void setROSReg( void)
{
	uint	ii, mask;

	if (ROSPortVal != ROSPortOld)
	{	OutPortVal &= 0xF8;									// reset data, clock and latch lines
		mask = ROSPortVal;									// get current port value

		for (ii = 0; ii < 6; ii++)
		{
			// place bit-0 of 'mask' into OUT-0 bit

			OutPortVal = (OutPortVal & 0xFE) | (mask & 0x01);

			outp( Port, OutPortVal);						// output bit
			OutPortVal |= 0x02;								// clock in data
			outp( Port, OutPortVal);						// update port
			OutPortVal &= 0xFD;								// reset clock
			outp( Port, OutPortVal);						// update port
			mask >>= 1;											// get next bit
		}
		OutPortVal |= 0x04;									// latch in new data
		outp( Port, OutPortVal);							// update port
		OutPortVal &= 0xFB;									// reset latch
		outp( Port, OutPortVal);							// update port
		ROSPortOld = ROSPortVal;							// remember new value
		
		OutPortOld = OutPortVal;							// remember new Port value
	}
}			

// Set ROSbit to value in 'val'.

void setROSBit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	ROSPortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in port
	ROSPortVal |= val;										// set/reset bit(s) in port
}

// Read an ROSbit

uint readROSBit( Sound_s *snd)
{
	return (((ROSPortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}

/***** Solar Quest 8 bit shift register *****/

// Set the 8 bit shift register

void setSQ1Reg( void)
{
	uint	ii, mask;

	if (SQ1PortVal != SQ1PortOld)
	{	OutPortVal &= 0x6D;									// reset data, clock and latch lines
		mask = SQ1PortVal;									// get current port value

		for (ii = 0; ii < 8; ii++)
		{
			// place bit-0 of 'mask' into OUT-7 bit

			if (mask & 0x01)
				OutPortVal |= 0x80;

			else
				OutPortVal &= 0x7F;

			outp( Port, OutPortVal);						// output bit
			OutPortVal |= 0x10;								// clock in data
			outp( Port, OutPortVal);						// update port
			OutPortVal &= 0xEF;								// reset clock
			outp( Port, OutPortVal);						// update port
			mask >>= 1;											// get next bit
		}
		OutPortVal |= 0x02;									// latch in new data
		outp( Port, OutPortVal);							// update port
		OutPortVal &= 0xFD;									// reset latch
		outp( Port, OutPortVal);							// update port
		SQ1PortOld = SQ1PortVal;							// remember new value
		
		OutPortOld = OutPortVal;							// remember new Port value
	}
}			

// Set SQ1bit to value in 'val'.

void setSQ1Bit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	SQ1PortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in port
	SQ1PortVal |= val;										// set/reset bit(s) in port
}

// Read an SQ1bit

uint readSQ1Bit( Sound_s *snd)
{
	return (((SQ1PortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}


/***** Solar Quest 16 bit shift register *****/

// Set the 16 bit shift register

void setSQ2Reg( void)
{
	uint	ii, mask;

	if (SQ2PortVal != SQ2PortOld)
	{	OutPortVal &= 0x6E;									// reset data, clock and latch lines
		mask = SQ2PortVal;									// get current port value

		for (ii = 0; ii < 16; ii++)
		{
			// place bit-0 of 'mask' into OUT-7 bit

			if (mask & 0x01)
				OutPortVal |= 0x80;

			else
				OutPortVal &= 0x7F;

			outp( Port, OutPortVal);						// output bit
			OutPortVal |= 0x10;								// clock in data
			outp( Port, OutPortVal);						// update port
			OutPortVal &= 0xEF;								// reset clock
			outp( Port, OutPortVal);						// update port
			mask >>= 1;											// get next bit
		}
		OutPortVal |= 0x01;									// latch in new data
		outp( Port, OutPortVal);							// update port
		OutPortVal &= 0xFE;									// reset latch
		outp( Port, OutPortVal);							// update port
		SQ2PortOld = SQ2PortVal;							// remember new value
		
		OutPortOld = OutPortVal;							// remember new Port value
	}
}			

// Set SQ2bit to value in 'val'.

void setSQ2Bit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	SQ2PortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in port
	SQ2PortVal |= val;										// set/reset bit(s) in port
}

// Read an SQ2bit

uint readSQ2Bit( Sound_s *snd)
{
	return (((SQ2PortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}


/***** Tailgunner 3 to 8 mux *****/

// Set the Tailgunner register

void setTGSReg( void)
{
	uchar	ii, mask, setMask;

	setMask = TGSPortVal ^ TGSPortOld;				// get bits that need updating
	mask = TGSPortVal;					 				// get copy of port

	// loop through each bit setting the ones that need changing

	for (ii = 0; setMask != 0; ii++)
	{
		if (setMask & 0x01)
		{	OutPortVal &= 0xE0;							// remove old data, strobe and sel lines
			OutPortVal |= ii;								// set select lines
			OutPortVal |= (mask & 0x01) << 3;		// set data line
			OutPortVal |= 0x10;							// set strobe line high
			outp( Port, OutPortVal);					// place data and select lines on port
			OutPortVal &= 0xEF;							// set strobe line low
			outp( Port, OutPortVal);					// place data and select lines on port
			OutPortVal |= 0x10;							// set strobe line high
			outp( Port, OutPortVal);					// place data and select lines on port
		}
		setMask >>= 1;										// shift SET flag
		mask >>= 1;											// shift value mask
	}
	TGSPortOld = TGSPortVal;							// save new value
	
	OutPortOld = OutPortVal;							// remember new Port value
}			

// Set TGSbit to value in 'val'.

void setTGSBit( Sound_s *snd, uint val)
{
	val &= snd->bitMask;										// turn off all but needed bit(s)
	val ^= snd->invMask;										// invert bit(s) if needed
	val <<= snd->shlCount;									// shift value into position
	TGSPortVal &= ~(snd->bitMask << snd->shlCount);	// turn off bit(s) in reg
	TGSPortVal |= val;										// set/reset bit(s) in reg
}

// Read a TGSbit

uint readTGSBit( Sound_s *snd)
{
	return (((TGSPortVal >> snd->shlCount) ^ snd->invMask) & snd->bitMask);
}

void (*SetRegTbl[])( void) =
{	setOutReg,
	setAASReg,
	setROSReg,
	setSQ1Reg,
	setSQ2Reg,
	setTGSReg
};

void (*SetBitTbl[])( Sound_s *snd, uint val) =
{	setOutBit,
	setAASBit,
	setROSBit,
	setSQ1Bit,
	setSQ2Bit,
	setTGSBit
};

uint (*ReadBitTbl[])( Sound_s *snd) =
{	readOutBit,
	readAASBit,
	readROSBit,
	readSQ1Bit,
	readSQ2Bit,
	readTGSBit
};	

// These are the universal routines called to set a 'register' or set a 'bit'
// or read a bit in a register.  They do this by indexing the routines setup
// in the above tables.

// After the proper bits have been setup, this routine moves the data from
// an internal copy of the register to the hardware register on the sound
// board.

void setReg( Sound_s *snd)
{
	SetRegTbl[snd->type]();
}

// Set the value of a register's bit in the internal copy of the register.
// This routine alone, does not update the sound card hardware.  After
// using this routine, 'setReg()' is called to actually update the hardware.

void setBit( Sound_s *snd, uint val)
{
	SetBitTbl[snd->type]( snd, val);
}

// Read the value of a register's bit from the internal copy of the register.
// Since the hardware registers of the sound boards are all write only, this
// routine does can not (and does not) read from the actual hardware.

uint readBit( Sound_s *snd)
{
	return (ReadBitTbl[snd->type]( snd));
}

// Calculate the cursor position of sound description.

void	getSndPos( uint num, uint *xx, uint *yy)
{
	if (num >= SNDROWS)
		*xx = COLUMN2;

	else
		*xx = COLUMN1;

	*yy = (num % SNDROWS) + 4;
}

// Initialize a sound board

void initSoundBd( SoundBd_s *sbd)
{
	uint		ii, xx, yy, cc;
	Sound_s	*snd;												// pointer to 'sounds' array

	snd = sbd->sound;											// point to sounds

	// initialize main port values before initializing any sub ports

	for (ii = 0; ii < sbd->nsnd; ii++)
		if (snd[ii].type == ST_OUT)
			setBit( snd + ii, 0);							// turn off sound bit

	// set main (ST_OUT) port to its initial value

	OutPortOld = ~OutPortVal;								// make sure register is set

	for (ii = 0; ii < sbd->nsnd; ii++)
		if (snd[ii].type == ST_OUT)
			setReg( snd + ii);								// set port

	// initialize the bits of all sub (anything that's not an ST_OUT) ports

	for (ii = 0; ii < sbd->nsnd; ii++)
		if (snd[ii].type != ST_OUT)
			setBit( snd + ii, 0);							// turn off sound bit

	// set any sub ports (shift register, etc.) to their initial values

	// ***** All registers needing initialization should be pre-set
	// ***** here to always force a full register initialization.

	AASPortOld = ~AASPortVal;								// make sure AAS register is set
	ROSPortOld = ~ROSPortVal;								// make sure ROS register is set
	SQ1PortOld = ~SQ1PortVal;								// make sure SQ1 register is set
	SQ2PortOld = ~SQ2PortVal;								// make sure SQ2 register is set
	TGSPortOld = ~TGSPortVal;								// make sure TGS register is set

	for (ii = 0; ii < sbd->nsnd; ii++)
		if (snd[ii].type != ST_OUT)
			setReg( snd + ii);								// set all sub ports

	// initialize screen

	ii = strlen( Sbd->title);
	ii += ii - 1;
	ii = (80 - ii) / 2;
	gotoxy( ii, 2);

	// print sound board's name as all capitals with spaces.

	for (ii = 0; Sbd->title[ii] != 0; ii++)
	{	putch( toupper( Sbd->title[ii]));
		putch( ' ');
	}

	for (ii = 0; ii < Sbd->nsnd; ii++)
	{	getSndPos( ii, &xx, &yy);
		gotoxy( xx, yy);

		// convert menu number to base 36 (0-Z)

		cc = ii + '0';

		if (cc > '9')
			cc += 'A' - '9';

		cprintf( "%c) [   0] %s", cc, snd[ii].name);
	}
}

// Return to DOS after re-initializing sound board (turn off all sounds), 
// restoring the 8253/8254 timer to it's default mode, and sending the
// cursor to the bottom of the screen.

void exit2dos( uint err)
{	initSoundBd( Sbd);
	gotoxy( 1, 25);
	fputc( '\n', stdout);
	restore_8253();
	exit( err);
}

// Convert a character, '0'-'Z' to a menu value, will exit to DOS
// if an ESC character is found.

uint cc2menu( uchar cc)
{
	uint	menu;

	if (cc == '\x1B')
		exit2dos( 0);

	cc = toupper( cc);

	if (cc >= '0' && cc <= '9')
		menu = cc - '0';

	else if (cc >= 'A' && cc <= 'Z')
		menu = cc - 'A' + 0x0A;

	else
		menu = (uint)-1;

	return (menu);
}

// Wait for a valid menu key to be pressed, return it's menu value.

uint getmenu( uint max)
{
	uint	menu;

	do
	{	menu = cc2menu( getch());
	} while (menu >= max);

	return (menu);
}

// Get a decimal value from screen.
// If 'max' is less than 10, then no waiting (or editing) of value is allowed.

uint getval( uint max)
{
	uint	val, xx, yy, ccCount;
	uint	cc;

	xx = wherex();
	yy = wherey();

	if (max < 10)
	{
		do
		{	cc = getch();

			if (cc == '\r')
				return (max + 1);

			val = cc2menu( cc);
		} while (val > max);

		ccCount = 1;
		putch( cc);
	}
	else
	{	val = 0;
		ccCount = 0;

		while (1)
		{	cc = getch();

			if (cc == '\x1B')
				exit2dos( 0);

			else if (cc == '\r')
				break;

			else if (cc == '\b')
			{	val /= 10;

				if (ccCount > 0)
					ccCount--;					// update character count
			}

			if (cc >= '0' && cc <= '9')
			{
				cc = val * 10 + (cc - '0');

				if (cc <= max)
				{	val = cc;

					if (ccCount == 0 || val > 9)
						ccCount++;				// update character count
				}
			}

			gotoxy( xx, yy);

			if (ccCount != 0)
				cprintf( "%u", val);

			clreol();
		}
	}

	if (ccCount == 0)
		return (max + 1);

	return (val);
}

// Trigger a sound

void trgSound( uint snum, uint action, uint val, Sound_s *snd)
{
	uint	ii, xx, yy;
	uint	jj, sndState;

	// point to sound

	snd += snum;

	switch (action)
	{
	case ACT_TGL:
		val = readBit( snd) ^ 1;						// invert value
		setBit( snd, val);
		setReg( snd);

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cprintf( "%4u", val);
		break;

	case ACT_PLS:
		setBit( snd, 1);									// turn on sound bit
		setReg( snd);										// turn on sound

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cputs( "   1");

		if (snd->timing != 0)
			for (ii = 0; ii < snd->timing; ii++)
				waittimer_qk( PULSE38HZ);				// wait for sound pulse length

		setBit( snd, 0);									// turn off sound bit
		setReg( snd);										// turn off sound

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cputs( "   0");
		break;

	// Special Tailgunner sound hack.  This is used to trigger the laser sound
	// in the same way the game does.  Just a little nicety to make the game
	// card 'sound' right when it's fuctioning properly.  (Not to mention how
	// this helps in the capturing of sound samples for emulators.)

	case ACT_TGR:
		sndState = 1;

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cputs( "   1");

		// pulse sound twice, leaving it turned on

		for (jj = 0; jj < 5; jj++)
		{	setBit( snd, sndState);
			setReg( snd);
			sndState ^= 1;
			waittimer_qk( PULSE38HZ);					// wait for sound pulse length
		}

		// wait for final pulse time, then turn off sound

		if (snd->timing != 0)
			for (ii = 0; ii < snd->timing; ii++)
				waittimer_qk( PULSE38HZ);				// wait for sound pulse length

		setBit( snd, 0);									// turn off sound bit
		setReg( snd);										// turn off sound

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cputs( "   0");
		break;

	// Default to 'ACT_SET'

	default:
		setBit( snd, val);								// set sound value
		setReg( snd);										// set port

		getSndPos( snum, &xx, &yy);
		gotoxy( xx + VALOFS, yy);
		cprintf( "%4u", val);
		break;
	}
}

void spam( void)
{
	// tell user how to exit program

	gotoxy( 27, 25);
	cputs( "Press <ESC> to exit to DOS");

	// Inflate author's ego by spamming user with authors name.

	gotoxy( 22, 18);
	cputs( "Cinematronics Sound Board Exercisor");
	gotoxy( 34, 20);
	cputs( "Version 0.1");
	gotoxy( 29, 22);
	cputs( "Written by Zonn Moore");
}

void main( int argc, char *argv[])
{
	uint	cc, yy, menu, override, val, numSbds;

	if (argc <= 1)
	{
		// As a future addition an auto port search routine could be added here
		// by looking for a port with sometype of loopback wires added to the parallel
		// port cable.  For now just use the predefined default port value if one is not
		// given.

		Port = DEFPORTADR;		// (possibly add an auto port search routine later)

		// Indicate how the port could be changed on the command line, then use default.

		fputs( "\nUse: CSXOR port_adr\n", stderr);
		fputs( "\nWhere:", stderr);
		fputs( "\n   port_adr = Address of parallel port (in hex).\n", stderr);
		fprintf( stderr, "\nUsing the default port address: %03X\n", Port);
		fputs( "\nPress any key to continue...", stderr);
		getch();
	}
	else
	{
		// Read port address from command line as a hex address

		Port = (uint)strtol( argv[1], 0, 16);
		fprintf( stderr, "\nUsing port address: %03X\n", Port);
		getch();
	}

	init_8253();

	// From this point on use built in Borland C library functions for
	// writing to the video screen.  This allows absolute addressing of
	// the cursor, etc.  This could also be done with ANSI commands 
	// (with slightly more effort) to make this code more portable to
	// other C compilers.

	clrscr();						// clear screen
	spam();							// info display

	gotoxy( 19, 2);				// set cursor position
	cputs( "S O U N D   B O A R D   S E L E C T I O N");

	// Print a list of available sound cards

	for (numSbds = 0; numSbds < NUMMEM( SoundBoard); numSbds++)
	{	gotoxy( COLUMN1, numSbds + 4);

		// convert to base 36 (0-Z)

		cc = numSbds + '0';

		if (cc > '9')
			cc += 'A' - '9';

		cprintf( "%c) %s", cc, SoundBoard[numSbds].title);
	}

	// ask for a card

	gotoxy( COLUMN1, numSbds + 5);
	cputs( "Choose a sound board: ");

	menu = getmenu( numSbds);	// get sound board offset
	Sbd = SoundBoard + menu;	// point to sound board

	// initialize sound board, also prints sound menu

	clrscr();
	spam();							// info display
	initSoundBd( Sbd);

	// find Y position of inquiry line

	if (Sbd->nsnd >= SNDROWS)
		yy = SNDROWS + 5;

	else
		yy = Sbd->nsnd + 5;

	// start asking for sound commands

	while (1)
	{	gotoxy( COLUMN1, yy + 1);
		clreol();
		gotoxy( COLUMN1, yy);
		cputs( "Choose sound, or 'T' to Toggle, 'S' to Set, or 'I' to Initialize:  \b");

		override = 0;

		do
		{
			cc = getch();					// wait for a character

			while (kbhit())
				cc = getch();				// don't let type ahead buffer fill up, use last char pressed

			// <ESC> to exit program

			if (cc == '\x1B')
				exit2dos( 0);

			cc = toupper( cc);

			if (cc == 'I')					// re-init board?
				break;

			if (cc == 'T')
				override = ACT_TGL;
			
			else if (cc == 'S')
				override = ACT_SET;

			if (override != 0)
			{	gotoxy( COLUMN1, yy);
				clreol();

				if (override == ACT_TGL)
					cputs( "Toggle sound: ");

				else if (override == ACT_SET)
					cputs( "Set sound: ");

				do
				{	cc = getch();
					menu = cc2menu( cc);
				} while (menu >= Sbd->nsnd);
			}
			else
				menu = cc2menu( cc);

		} while (menu >= Sbd->nsnd);

		putch( cc);

		if (cc == 'I')
			initSoundBd( Sbd);

		else
		{	if (override == 0)
				override = Sbd->sound[menu].action;

			if (override == ACT_SET)
			{	gotoxy( COLUMN1, yy + 1);
				cputs( "Enter new value: ");
				val = getval( Sbd->sound[menu].bitMask);

				// if value is out of range, don't change value

				if (val > Sbd->sound[menu].bitMask)
					continue;			// jump top of loop
			}
			trgSound( menu, override, val, Sbd->sound);
		}
	}
}
